/*
==============================================================================
Title: "Map Attribute Changer"
(c) 1999 by Matthias "Maddes" Buecher

Authors:
Matthias "Maddes" Buecher <maddes@bigfoot.com>

Idea:
Daniel "Armadillo" Roberts <armadillo@fidra30.freeserve.co.uk>

Additional credits:
Marc Stern <stern@mble.philips.be>
  Function strmvstr() Copyright (C) 1993

Notes:
This program was created to provide a tool to use
WorldCraft 2.0 for Classic Quake map editing.

Most of this code is easy to comprehend and there should not be too many
questions about how it works.  The heart of the program has been slighty
commented out to help you understand the technical operations being done.

Important:
It should compile under Win32 and Unix with no errors, some warnings may still occur.
If you use a processor which has another byte order than Intel (lowest to highest byte),
you have to reprogram all offset functions, etc.

DISCLAIMER WARNING:
WE DO NOT ACCEPT RESPONSIBILITY FOR ANY EFFECTS, POSITIVE OR NEGATIVE,
THAT THIS CODE MAY CAUSE TOWARDS YOUR COMPUTER.
WE DO NOT SUPPORT MODIFICATIONS OF THIS CODE, USE ONLY AT YOUR OWN RISK.
THIS PATCH HAS NOT BEEN ENDORSED BY THE ID SOFTWARE CORPORATION.
THIS PROGRAM IS FREE OF CHARGE. IF YOU HAVE RECEIVED THIS PROGRAM THROUGH
PAYMENT, PLEASE CONTACT US.
THANK YOU.
==============================================================================
*/

/* Compatible headers */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

/* needed to compile under Linux */
#ifndef __WIN32__
#define stricmp strcasecmp
#endif

/* constants */
#define VERSION "1.00 (1999-11-18)"

#define BUFLEN 1024
#define STRLEN 128

#define PARM_CHAR	'-'

#define COMMENT_CHAR	';'

#define ATTR_BEGIN_CHAR	'['
#define ATTR_END_CHAR	']'

/*
structure definitions of lists for attributes -> texts
*/
typedef struct s_textentry {
	char oldtext[STRLEN], newtext[STRLEN];
	int oldlength, newlength;
	struct s_textentry *next;
} textentry;

typedef struct s_attrentry {
	char name[STRLEN];
	int length;
	struct s_textentry *texts;
	struct s_attrentry *next;
} attrentry;

/* global data definition */
char *default_ini_file = "mapchanger.ini";	/* pointer on default ini file name */
char *ini_file = NULL;				/* pointer on used ini file name */
FILE *ini;					/* ini file handle */

char *data_file = NULL;				/* pointer on data file name */
FILE *data;					/* data file handle */

char tmp_file[BUFLEN];				/* temp file name */
FILE *tmp;					/* temp file handle */

char cnv_file[BUFLEN];				/* converted file name */

char buffer1[2*BUFLEN];				/* read buffer */
char buffer2[2*BUFLEN];				/* compare buffer */

int flag_error;					/* error occured */
int flag_warning;				/* warning occured */

int flag_undo;					/* undo flag */
int flag_yes;					/* assume yes flag */
int flag_test;					/* test flag */

/*
******************************************************************************
*
*   G E N E R A L   F U N C T I O N S
*
******************************************************************************
*/

/***
 *
 *  Found on http://www.programmersheaven.com/ and changed to incasesensitive
 *
 *  Function    :  strmvstr
 *
 *  Copyright (C) 1993   Marc Stern  (internet: stern@mble.philips.be)
 *
 *  Description :  Copy an input string in an output string
 *		   with replacing all occurences of a find_str string.
 *
 *  Parameters  :  out      char  *out_str      result
 *                 in       char  *in_str       in string
 *                 in       char  *find_str     find_str string to replace
 *                 in       char  *replace_str  string to put in place of find_str
 *
 *  Return code :  pointer to result.
 *
 *  OS/Compiler :  All
 ***/

char *strimvstr( char *out_str, const char *in_str, const char *find_str, const char *replace_str )
{
	char *ptr = out_str;					/* store return value */
	const char *ptr_new = replace_str;			/* store pointer to replace_str */
	unsigned int count;

	count = 0;
	for ( ; *out_str = *in_str; out_str++, in_str++ )	/* this copies the in_str to the out_str */
	{
		if ( tolower(*out_str) == tolower(*find_str) )	/* copied char matches char of find_str? */
		{
			find_str++;				/* next char in find_str */
			count++;				/* increase count for later reset */

			if ( ! *find_str )			/* find_str ended? = fully matched find_str */
			{
				out_str -= count - 1;		/* back to first char match of find_str */
				find_str -= count;		/* reset find_str */
				count = 0;			/* reset counter */
				while ( *out_str++ = *replace_str++ );	/* copy replace_str */
				replace_str = ptr_new;		/* reset replace_str, not handled with count */
				out_str --; out_str --;		/* subtract end char and recognize increment of for() */
			}
		}
		else if (count)					/* not fully matched, reset pointers */
		{
			find_str -= count;			/* reset find_str */
			out_str -= count;
			in_str -= count;
			count = 0;
		}
	}

	return ptr;
}

/*
==============================================================================
search string within string, adapted from strimvstr()
==============================================================================
*/
char *stristr( char *in_str, const char *find_str)
{
	unsigned int count;

	count = 0;
	for ( ; *in_str ; in_str++ )
	{
		if ( tolower(*in_str) == tolower(*find_str) )			/* char matches char of find_str? */
		{
			find_str++;				/* next char in find_str */
			count++;				/* increase count for later reset */

			if ( ! *find_str )			/* find_str ended? = fully matched find_str */
			{
				in_str -= count - 1;		/* back to first char match of find_str */
				return in_str;
			}
		}
		else if (count)					/* not fully matched, reset pointers */
		{
			find_str -= count;			/* reset find_str */
			in_str -= count;
			count = 0;
		}
	}

	return NULL;
}

/*
==============================================================================
displays the given text and asks for confirmation
==============================================================================
*/
int confirm(char *question)
{
	char answer;

	printf("%s [Y/N] ", question);

	do
	{
		if (flag_yes)
		{
			answer = 'Y';
			printf("Y\n");
		}
		else
			answer=toupper(getchar());
	}
	while ((answer != 'Y') && (answer != 'N'));	/* loop until valid answer */

	if (answer == 'Y')
		return 1;
	else
		return 0;
}

/*
==============================================================================
trims the string from all leading spaces/tabs
and all trailing spaces/tabs/carriage return
returns the new length
==============================================================================
*/
int strtrim(char *text)
{
	int i;
	int length;

	/* leading spaces */
	length = strlen(text);
	for (i=0; (i<length) && ((text[i]==' ') || (text[i]=='\t') || (text[i]==(char)0x0A) || (text[i]==(char)0x0D)); i++);
	if (i)
		strcpy(text, text+i);

	/* trailing spaces */
	length = strlen(text);
	for (i=length-1; (i>=0) && ((text[i]=='\n') || (text[i]==' ') || (text[i]=='\t') || (text[i]==(char)0x0A) || (text[i]==(char)0x0D)); i--);
	i++;
	text[i] = 0;

	return i;
}

/*
==============================================================================
transforms a string to lowercase
==============================================================================
*/
void strlower(char *text)
{
	int i;
	int length;

	length = strlen(text);
	for (i=0; i<length; i++)
		text[i] = tolower(text[i]);
}

/*
==============================================================================
gets the path out of a filename in length
==============================================================================
*/
int get_pathlength(const char *const filename)
{
	const char *colon;
	const char *slash;
	const char *backslash;

	colon = strchr(filename, ':');
	slash = strrchr(filename, '/');
	backslash = strrchr(filename, '\\');

	if (slash > backslash)
	{
		slash++;
		return (slash - filename);
	}
	else if (backslash > slash)
	{
		backslash++;
		return (backslash - filename);
	}
	else if (colon)
	{
		colon++;
		return (colon - filename);
	}

	return 0;	// no path found
}

/*
==============================================================================
gets the filename out of the path
==============================================================================
*/
char *get_filename(char *const filename)
{
	char *colon;
	char *slash;
	char *backslash;

	colon = strchr(filename, ':');
	slash = strrchr(filename, '/');
	backslash = strrchr(filename, '\\');

	if (slash > backslash)
	{
		slash++;		// first char after slash
		return slash;
	}
	else if (backslash > slash)
	{
		backslash++;		// first char after backslash
		return backslash;
	}
	else if (colon)
	{
		colon++;		// first char after colon
		return colon;
	}

	return filename;	// no path found
}

/*
==============================================================================
gets the extension out of the path
==============================================================================
*/
char *get_extension(const char *const filename)
{
	char *dot;
	char *slash;
	char *backslash;

	dot = strrchr(filename, '.');
	slash = strrchr(filename, '/');
	backslash = strrchr(filename, '\\');

	if ((dot) && (dot > slash) && (dot > backslash))
	{
		dot++;		// first char after dot
		return dot;
	}

	return NULL;		// no extension found
}

/*
==============================================================================
frees all the memory used
==============================================================================
*/
void freemem(attrentry *attr)
{
	attrentry *delattr;
	textentry *text;
	textentry *deltext;

	while (attr)
	{
		text = attr->texts;
		while (text)
		{
			deltext = text;
			text = text->next;
			free(deltext);
		}
		delattr = attr;
		attr = attr->next;
		free(delattr);
	}
}

/*
******************************************************************************
*
*   U S E R   I N T E R F A C E
*
******************************************************************************
*/

/*
==============================================================================
displays the help text
==============================================================================
*/
void display_help(char *call)
{
	/* syntax */
	printf("Usage: MapChanger [-?] [-t] [-u] [-y] [-i <ini file>] file\n");
	printf("  -?           : displays this help text\n");
	printf("  -t           : testmode\n");
	printf("  -u           : undo previous changes done with the ini file\n");
	printf("  -y           : assumes yes on all queries\n");
	printf("  -i <ini file>: defines the ini file to be used\n");
	printf("                 default is \"%s\"\n", default_ini_file);

	/* example */
	printf("\nExample:\n");
	if (data_file)
	{
		printf("%s -u \"%s\"\n", call, data_file);
	}
	else
	{
		printf("%s -u \"example.map\"\n", call);
	}
	printf("This would undo all previous changes done with the default ini file\n");

	/* contact */
	printf("\nHow to contact the author:\n");
	printf("Matthias \"Maddes\" Buecher <maddes@bigfoot.com>\n");

	printf("\nThis program was created to provide a tool to use WorldCraft 2.0 for Classic Quake map editing.\n");
}

/*
==============================================================================
checks if string begins with a parameter character
==============================================================================
*/
int chk_paramchar(char *text)
{
	/* check for parameter character at beginning */
	if (text[0] == PARM_CHAR)
	{
		strlower(text);
		return 1;
	}
	else
		return 0;
}

/*
==============================================================================
checks all parameters
when help parameter is found it displays the help text and leaves the function
checks if the ini file is present
displays all used parameters and asks for confirmation
==============================================================================
*/
int chk_param(int argc, char *argv[])
{
	/* return values: 0 = confirmed, 1 = help called, 2 = error in parameters, 3 = declined confirmation */

	/* define local variables */
	int i;			/* iteration variable */

	int dif;		/* difference between statement and parameter number, used for correct output */

	int flag_help;		/* display help */
	int flag_inifile;	/* ini file stated in path */

	char *filename;		/* pointer to file name */

	/* initialize local variables */
	dif = 0;
	flag_error = 0;
	flag_warning = 0;
	flag_help = 0;
	flag_inifile = 1;

	/* parse parameters */
	for (i=1; i<argc; i++)
	{
		if (chk_paramchar(argv[i]))
		{
			if (!strcmp(argv[i],"-?") || !strcmp(argv[i],"-h") || !strcmp(argv[i],"-help"))	/* help parameter */
			{
				flag_help = 1;
				flag_error = 1;
				break;
			}
			else if (!strcmp(argv[i],"-t") || !strcmp(argv[i],"-test"))	/* test parameter */
			{
				/* check already set */
				if (flag_test)
				{
					printf("Warning on parameter %i: test already defined\n", i-dif);
					flag_warning = 1;
				}

				flag_test=1;
			}
			else if (!strcmp(argv[i],"-y"))	/* assume yes on all queries parameter */
			{
				/* check already set */
				if (flag_yes)
				{
					printf("Warning on parameter %i: assume yes already defined\n", i-dif);
					flag_warning = 1;
				}

				flag_yes=1;
			}
			else if (!strcmp(argv[i],"-u"))	/* undo parameter */
			{
				/* check already set */
				if (flag_undo)
				{
					printf("Warning on parameter %i: undo already defined\n", i-dif);
					flag_warning = 1;
				}

				flag_undo=1;
			}
			else if (!strcmp(argv[i],"-i"))	/* ini file parameter */
			{
				i++;	/* next statement (normally filename) */

				if ( (i<argc) && (!chk_paramchar(argv[i])) )	/* statement available and not a *new parameter* */
				{
					/* as this parameter has two statements, decrease the parameter number by one */
					dif++;

					/* check already set */
					if (ini_file != default_ini_file)
					{
						printf("Warning on parameter %i: a custom ini file was already named\n", i-dif);
						flag_warning = 1;
					}

					/* set pointer of ini file name to statement string */
					ini_file = argv[i];

					/* check if filename include */
					flag_inifile = 1;
					filename = get_filename(ini_file);
					if (!*filename)
					{
						printf("Error on parameter %i: no file stated in path\n", i-dif);
						flag_inifile = 0;
						flag_help = 1;
						flag_error = 2;
					}
				}
				else
				{
					i--;

					printf("Error on parameter %i: custom ini file not named\n", i-dif);
					flag_help = 1;
					flag_error = 2;
				}
			}
			else				/* unknown parameter */
			{
				printf("Error on parameter %i: \"%s\" is an unknown parameter\n", i-dif, argv[i]);
				flag_help = 1;
				flag_error = 2;
			}
		}
		else	/* should be the filename */
		{
			if (data_file)
			{
				printf("Error on parameter %i: \"%s\", a data file was already named \"%s\"\n", i-dif, argv[i], data_file);
				flag_help = 1;
				flag_error = 2;
			}
			else
			{
				/* set pointer of ini file name to statement string */
				data_file = argv[i];
			}
		}
	}

	/* check if ini file exists */
	if (flag_inifile)
	{
		ini = fopen(ini_file,"r");
		if (!ini)
		{
			printf("Error: could not open ini file \"%s\"\n", ini_file);
			flag_warning = 1;

			// check for ini file in program directory
			buffer2[0] = 0;
			i = get_pathlength(argv[0]);
			if (i)
			{
				strncpy(buffer2, argv[0], i);
				buffer2[i] = 0;
			}
			strcat(buffer2, get_filename(ini_file));
			ini_file = buffer2;

			printf("       ...using ini file \"%s\"\n", ini_file);

			ini = fopen(ini_file,"r");
			if (!ini)
			{
				printf("Error: could not open ini file \"%s\"\n", ini_file);
				flag_error = 2;
			}
		}
		if (ini)
		{
			fclose(ini);
			ini = NULL;
		}
	}

	/* check if data file named and exists */
	data = NULL;
	if (!data_file)
	{
		printf("Error: no data file named\n", data_file);
		flag_error = 2;
		flag_help = 1;
	}
	else
	{
		/* create other filenames */
		strcpy(cnv_file, data_file);
		strcat(cnv_file, ".cnv");
		strcpy(tmp_file, data_file);
		strcat(tmp_file, ".tmp");

		/* remove old converted file if exists */
		data = fopen(cnv_file, "r");
		if (data)
		{
			fclose(data);

			if (!flag_test)
			{
				if (remove(cnv_file))
				{
					printf("Error: could not remove old converted file \"%s\"\n", cnv_file);
					flag_error = 2;
				}
				else
				{
					printf("Warning: deleted already existing file \"%s\"\n", cnv_file);
					flag_warning = 1;
				}
			}
		}

		data = fopen(data_file,"r");
		if (!data)
		{
			printf("Error: could not open data file \"%s\"\n", data_file);
			flag_error = 2;
		}
	}
	if (data)
	{
		fclose(data);
		data = NULL;
	}

	/* asked for help */
	if (flag_help)
	{
		if (flag_error > 1 || flag_warning)
		{
			printf("\n");
		}
		display_help(argv[0]);
	}
	/* any errors occured? */
	if (flag_error) return 1;

	/* display all data */
	if (flag_warning) printf("\n");
	printf("The following settings will be used:\n------------------------------------\n");
	printf("INI file: \"%s\"\n", ini_file);
	printf("Data file: \"%s\"\n", data_file);
	if (flag_undo) printf("UNDO enabled\n");
	if (flag_test) printf("TESTMODE enabled: \"%s\" will not be created\n", cnv_file);

	/* ask for confirmation and return */
	if (!confirm("\nDo you wish to proceed?"))
		return 1;

	printf("\n");

	return 0;
}

/*
******************************************************************************
*
*   I N I   F I L E   R E A D I N G   &   P A R S I N G
*
******************************************************************************
*/

/*
==============================================================================
reads the next line of the file
checks if line did not exceed buffer
converts the line for Unix
removes any comment
trims the line
checks if not an empty line, otherwise reads next line
==============================================================================
*/
char *read_iniline(char *buffer, int buflen, FILE *ini, int *line_count)
{
	int length;
	char *quote;
	char *comment;

	while (fgets(buffer, buflen, ini))
	{
		(*line_count)++;
		length = strlen(buffer);
		if ((length == buflen-1) && (buffer[length-1]!='\n'))
		{
			printf("Error in ini line %i: line too long, only %i characters allowed", *line_count, buflen-2);
			flag_error = 2;
			break;
		}

		if (buffer[0] == (char)0x0A || buffer[0] == (char)0x0D)		/* allow Unix and DOS/Windows files */
		{
			strcpy(buffer, &buffer[1]);
		}

		quote = buffer;
		comment = buffer;
		do
		{
			quote = strchr(quote, '\"');			/* search opening quote */
			comment = strchr(comment, COMMENT_CHAR);	/* search comment char */

			if (!quote || !comment)
				break;

			if (comment < quote)
				break;

			quote++;
			quote = strchr(quote, '\"');		/* search closing quote */

			if (!quote)				/* not closed, will be an error in parsing */
				break;

			quote++;
			comment = quote;
		} while (1);

		if (comment)
			comment[0] = 0;

		length = strtrim(buffer);		/* remove leading/trailing whitespaces */

		if (!length)
			continue;

		return buffer;
	}

	return NULL;
}

/*
==============================================================================
checks if line is a new attribute setting
removes the beginning and ending char
trims the line
==============================================================================
*/
int chk_attrline(char *line, int line_count)
{
	int i;

	if (line[0] == ATTR_BEGIN_CHAR)
	{
		i = strlen(line) - 1;

		if (line[i] != ATTR_END_CHAR)
		{
			printf("Error in ini line %i: attribute string not defined correctly", line_count);
			flag_error = 2;
		}
		else
		{
			line[0] = ' ';
			line[i] = 0;
			strtrim(line);
			return 1;
		}
	}

	return 0;
}

/*
==============================================================================
parses all data lines of the ini file
and creates lists out of the read data
==============================================================================
*/
attrentry *read_inifile(void)
{
	int line_count;		/* line number */

	int oldlength;
	int newlength;

	char *quote;

	char oldtext[BUFLEN];
	char newtext[BUFLEN];

	attrentry *start;
	attrentry *attr;
	attrentry *lastattr;
	textentry *text;
	textentry *lasttext;

	/* initialize local variables */
	start = NULL;
	attr = NULL;
	lastattr = NULL;
	lasttext = NULL;
	line_count = 0;
	flag_warning = 0;

	do			/* just to use break for errors */
	{
		/* open ini file */
		ini = fopen(ini_file, "r");
		if (!ini)
		{
			printf("Error: could not open ini file \"%s\"\n", ini_file);
			flag_error = 2;
			break;
		}

		/* reading ini file */
		while (!feof(ini) && read_iniline(buffer1, BUFLEN, ini, &line_count))
		{
			if (chk_attrline(buffer1, line_count))	/* attribute line */
			{
				if (flag_error) break;

				if (strlen(buffer1) > (STRLEN-1))
				{
					printf("Error on ini line %i: attribute too long, only %i characters allowed\n", line_count, STRLEN-1);
					flag_error = 2;
					break;
				}

				/* create attr list entry */
				attr = malloc(sizeof(attrentry));
				if (!attr)
				{
					printf("Error on ini line %i: couldn't allocate memory for new attribute\n", line_count);
					flag_error = 2;
					break;
				}
				strcpy(attr->name, buffer1);
				attr->length = strlen(attr->name);
				attr->texts = NULL;
				attr->next = NULL;

				if (!start) start = attr;
				if (lastattr) lastattr->next = attr;

				lastattr = attr;
				lasttext = NULL;
			}
			else if (attr)		/* text line */
			{
				/* first text entry */
				oldtext[0] = 0;
				if (buffer1[0] == '\"')
				{
					quote = strchr(&buffer1[1], '\"');	/* search closing quote */
					if (!quote)					/* not closed */
					{
						printf("Error on ini line %i: old text quotes not closed\n", line_count);
						flag_error = 2;
						break;
					}

					oldlength = quote - buffer1 - 1;
					strncpy(oldtext, &buffer1[1], oldlength);
					oldtext[oldlength] = 0;
					oldlength += 2;
				}
				else
				{
					buffer2[0] = 0;
					oldtext[0] = 0;
					sscanf(buffer1, "%s %s", oldtext, buffer2);
					oldlength = strlen(oldtext);
				}
				strcpy(buffer1, &buffer1[oldlength]);
				strtrim(buffer1);

				/* second text entry */
				newtext[0] = 0;
				if (buffer1[0] == '\"')
				{
					quote = strchr(&buffer1[1], '\"');	/* search closing quote */
					if (!quote)					/* not closed */
					{
						printf("Error on ini line %i: new text quotes not closed\n", line_count);
						flag_error = 2;
						break;
					}

					newlength = quote - buffer1 - 1;
					strncpy(newtext, &buffer1[1], newlength);
					newtext[newlength] = 0;
					newlength += 2;
				}
				else
				{
					buffer2[0] = 0;
					sscanf(buffer1, "%s %s", newtext, buffer2);
					newlength = strlen(newtext);
				}
				strcpy(buffer1, &buffer1[newlength]);
				strtrim(buffer1);

				oldlength = strlen(oldtext);
				newlength = strlen(newtext);

				if (oldlength > (STRLEN-1))
				{
					printf("Error on ini line %i: old text too long, only %i characters allowed\n", line_count, STRLEN-1);
					flag_error = 2;
					break;
				}
				if (newlength > (STRLEN-1))
				{
					printf("Error on ini line %i: new text too long, only %i characters allowed\n", line_count, STRLEN-1);
					flag_error = 2;
					break;
				}

				if (strlen(buffer1))
				{
					printf("Error on ini line %i: unexpected data following old and new text\n", line_count);
					flag_error = 2;
					break;
				}

				/* create text list entry */
				text = malloc(sizeof(textentry));
				if (!text)
				{
					printf("Error on ini line %i: couldn't allocate memory for text\n", line_count);
					flag_error = 2;
					break;
				}

				if (!flag_undo)
				{
					strcpy(text->oldtext, oldtext);
					strcpy(text->newtext, newtext);
					text->oldlength = oldlength;
					text->newlength = newlength;
				}
				else
				{
					strcpy(text->oldtext, newtext);
					strcpy(text->newtext, oldtext);
					text->oldlength = newlength;
					text->newlength = oldlength;

					if (!text->oldlength)		/* not possible without new text */
					{
						printf("Warning on ini line %i: can not undo without new text\n", line_count);
						flag_warning = 1;
						free(text);
						continue;
					}
				}
				text->next = NULL;

				if (lasttext) lasttext->next = text;
				else if (!attr->texts) attr->texts = text;
				else
				{
					printf("Error on ini line %i: text entry impossible, please contact the authors\n", line_count);
					flag_error = 2;
					break;
				}

				lasttext = text;
			}
			else
			{
				printf("Error on ini line %i: text defined before any attribute\n", line_count);
				flag_error = 2;
				break;
			}
		}

	} while (0);

	if (flag_error || flag_warning)
	{
		printf("\n");
	}

	if (ini)
	{
		fclose(ini);
		ini = NULL;
	}

	return start;
}

/*
******************************************************************************
*
*   F I L E S   R E A D I N G   &   W R I T I N G
*
******************************************************************************
*/

void process_file(attrentry *start)
{
	attrentry *attr;
	textentry *text;

	char *value;

	int length;
	int line_count;

	/* initialize local variables */
	line_count = 0;
	flag_warning = 0;

	do			/* just to use break for errors */
	{
		/* open data file */
		data = fopen(data_file, "r");
		if (!data)
		{
			printf("Error: could not open data file \"%s\"\n", data_file);
			flag_error = 2;
			break;
		}

		/* open tmp file */
		tmp = fopen(tmp_file, "w");
		if (!tmp)
		{
			printf("Error: could not create temp file \"%s\"\n", tmp_file);
			flag_error = 2;
			break;
		}

		/* reading data file */
		while (!feof(data) && fgets(buffer1, BUFLEN, data))
		{
			line_count++;
			length = strlen(buffer1);
			if ((length == BUFLEN-1) && (buffer1[length-1]!='\n'))
			{
				printf("Error in line %i: line too long, only %i characters allowed", line_count, BUFLEN-2);
				flag_error = 2;
				break;
			}

			attr = start;
			while (attr)
			{
				value = stristr(buffer1, attr->name);
				if (value)
				{
					strcpy(buffer2, buffer1);
					strtrim(buffer2);
					if (!(strncmp(buffer2, attr->name, attr->length)))
					{
						value += attr->length;
						printf("Found %s in line %i...\n", attr->name, line_count, value);
						flag_warning = 1;

						text = attr->texts;
						while (text)
						{
							if (stristr(value, text->oldtext))
							{
								printf("...replaced \"%s\" with \"%s\"\n", text->oldtext, text->newtext);
								strimvstr( buffer2, value, text->oldtext, text->newtext);
								strcpy(value, buffer2);
							}

							text = text->next;
						}
					}
				}

				attr = attr->next;
			}

			if (fputs(buffer1, tmp) == EOF)
			{
				printf("Error: couldn't write line %i to temp file \"%s\"\n", line_count, tmp_file);
				flag_error = 2;
				break;
			}
		}

	} while (0);

	if (flag_warning)
	{
		printf("\n");
	}

	if (data)
	{
		fclose(data);
		data = NULL;
	}

	if (tmp)
	{
		fclose(tmp);
		tmp = NULL;
	}

	if (!flag_error && !flag_test)
	{
		if (rename(tmp_file, cnv_file))
		{
			printf("Error: could not rename temp file \"%s\" to \"%s\"\n", tmp_file, cnv_file);
			flag_error = 2;
		}
	}
	else
	{
		remove(tmp_file);
	}
}

/*
******************************************************************************
*
*   M A I N   F U N C T I O N   ( S T A R T   O F   P R O G R A M )
*
******************************************************************************
*/

/*
==============================================================================
here the program starts and all sub functions are called
==============================================================================
*/
int main(int argc, char *argv[])
{
	int stop;

	attrentry *start;

	/* initialize variables */
	ini_file = default_ini_file;		/* pointer on default ini file name */
	start = NULL;
	flag_undo = 0;
	flag_yes = 0;
	flag_test = 0;

	/* display program header lines */
	printf("Map Attribute Changer\n");
	printf("(c) 1999 by Matthias \"Maddes\" Buecher\n");
	printf("Idea by Daniel \"Armadillo\" Roberts\n");
	printf("Version %s\n\n", VERSION);

	/* check parameters and let them confirm */
	stop = chk_param(argc, argv);
	if (stop)
	{
		exit(flag_error);
	}

	/* here we go with the "real" program */
	start = read_inifile();

	if (!flag_error)
	{
		process_file(start);
	}

	if (start) freemem(start);

	if (flag_error)
	{
		exit(flag_error);
	}

	printf("Finished\n");

	return 0;
}
